
import os
import symbolx as syx
from symbolx import DataCollection, SymbolsHandler, Symbol

print("Test 1")
folder = "csv_0"
syx.compress_subdirs(folder=folder, zip_extension='7z', delete=False)
# syx.unzip_all(folder=folder, zip_extension='7z', delete=False)
DC = DataCollection()
DC.add_collector(collector_name='csv_collector', parser=syx.symbol_parser_csv, loader=syx.load_csv)
DC.add_folder('csv_collector', folder)
DC.add_custom_attr(collector_name='csv_collector')
DC.adquire(id_integer=False, zip_extension='7z')
# DC.adquire(id_integer=False, zip_extension=None)
SH = SymbolsHandler(method='object', obj=DC)

var1 = Symbol(name='VAR1', symbol_handler=SH)
var2 = Symbol(name='VAR2', symbol_handler=SH)
var3 = Symbol(name='VAR3', symbol_handler=SH)

print(var1.dfc)
print(var2.dfc)
print(var3.dfc)

var1*var3
var2*var3

os.remove(os.path.join(folder,"s01.7z"))
os.remove(os.path.join(folder,"s02.7z"))

print("Test 2")
folder = "csv_1"
syx.compress_subdirs(folder=folder, zip_extension='7z', delete=False)
# syx.unzip_all(folder=folder, zip_extension='7z', delete=False)
DC = DataCollection()
DC.add_collector(collector_name='csv_collector', parser=syx.symbol_parser_csv, loader=syx.load_csv)
DC.add_folder('csv_collector', folder)
DC.add_custom_attr(collector_name='csv_collector')
DC.adquire(id_integer=False, zip_extension='7z')
# DC.adquire(id_integer=False, zip_extension=None)
SH = SymbolsHandler(method='object', obj=DC)

var1 = Symbol(name='VAR1', symbol_handler=SH)
var2 = Symbol(name='VAR2', symbol_handler=SH)
var3 = Symbol(name='VAR3', symbol_handler=SH)

var1*var3
var2*var3

os.remove(os.path.join(folder,"s01.7z"))
os.remove(os.path.join(folder,"s02.7z"))

print("Test 3")
folder = "feather_0"
syx.compress_subdirs(folder=folder, zip_extension='7z', delete=False)
# syx.unzip_all(folder=folder, zip_extension='7z', delete=False)
DC = DataCollection()
DC.add_collector(collector_name='feather_collector', parser=syx.symbol_parser_feather, loader=syx.load_feather)
DC.add_folder('feather_collector', folder)
DC.add_custom_attr(collector_name='feather_collector', with_='pandas')
DC.adquire(id_integer=False, zip_extension='7z')
# DC.adquire(id_integer=False, zip_extension=None)
SH = SymbolsHandler(method='object', obj=DC)

EnergyBalance = Symbol(name='EnergyBalance', value_type='m', symbol_handler=SH)

os.remove(os.path.join(folder,"Run_0001_20221220193231_1.7z"))
os.remove(os.path.join(folder,"Run_0002_20221220193338_1.7z"))


print("Test 4")
print("This requires GAMS API, otherwise, it will shoot an import error")
folder = "gdx_0"
syx.compress_subdirs(folder=folder, zip_extension='7z', delete=False)
# syx.unzip_all(folder=folder, zip_extension='7z', delete=False)
DC = DataCollection()
DC.add_collector(collector_name='gdx_collector', parser=syx.symbol_parser_gdx, loader=syx.load_gdx)
DC.add_folder('gdx_collector', folder)
DC.add_custom_attr(collector_name='gdx_collector', 
                    inf_to_zero=True, 
                    verbose=False)
DC.adquire(id_integer=True, gams_dir=os.getenv("GAMS_DIR"), zip_extension='7z')
# DC.adquire(id_integer=False, gams_dir=os.getenv("GAMS_DIR"), zip_extension=None)
SH = SymbolsHandler(method='object', obj=DC)


G_L = Symbol(name='G_L', symbol_handler=SH)


print(G_L.dfm)

os.remove(os.path.join(folder,"Run_0000_20221208200610_b000_r0000.7z"))
os.remove(os.path.join(folder,"Run_0001_20221208200610_b001_r0000.7z"))
os.remove(os.path.join(folder,"Run_0002_20221208200610_b002_r0000.7z"))
os.remove(os.path.join(folder,"Run_0003_20221208200610_b003_r0000.7z"))

print("Done!")


import re
from typing import Callable
from .utils import load_scenario_info, _get_files_path_list, select_scenario_files


class DataCollection:
    def __init__(self):
        '''
        Collect relevant data from raw scenarios output and organizes it per symbol or variable. No input is required. 
        This class defines a collector with custom parsers and loaders of the raw scenarios output.
        It also requires custom attributes that are the parser and loader input arguments.
        This structure helps to use personalized parsers and loaders depending on the type and format of the optimization tools' output.
        
        Examples
        --------
        Example 1: parser and loader from arrow tables and feather file format

            >>> import symbolx as syx
            >>> from symbolx import DataCollection
            >>> DC = DataCollection()
            >>> DC.add_collector(collector_name='opt_model', parser=syx.symbol_parser_feather, loader=syx.load_feather)
            >>> DC.add_folder(collector_name='opt_model', './raw_model_output')
            >>> DC.add_custom_attr(collector_name='opt_model', with_='pandas')
            >>> DC.adquire(id_integer=True, zip_extension=None)


        Example 2: parser and loader from GAMS GDX file format

            >>> import os
            >>> import symbolx as syx
            >>> from symbolx import DataCollection
            >>> DC = DataCollection()
            >>> DC.add_collector(collector_name='gdx_collector', parser=syx.symbol_parser_gdx, loader=syx.load_gdx)
            >>> DC.add_folder('gdx_collector', './gdx_output')
            >>> DC.add_custom_attr(collector_name='gdx_collector', inf_to_zero=True, verbose=False)
            >>> DC.adquire(id_integer=False, gams_dir=os.getenv("GAMS_DIR"), zip_extension=None)

        
        '''
        self.collector = {}
        self.scenarios_path = None
        self.config = None
        self.data = None
        self.symbol_name_list = None
        self.symbol_valuetype_dict = None
        self.short_names = None
        self.metadata_template = None
        self.scenarios_metadata = None
        self.symbols_book = None

    def add_collector(self, collector_name:str, parser:Callable, loader:Callable):
        '''
        Add a collector to the DataCollection object.

        Parameters
        ----------
        collector_name : str
            Define or set the name of the collector.
        parser : Callable
            The parser function.
        loader : Callable
            The loader function.
            
        Returns
        -------
        None.

        '''

        self.collector[collector_name] = {}
        self.collector[collector_name]['parser'] = parser
        self.collector[collector_name]['loader'] = loader
        self.add_symbol_list(collector_name)
        self.add_custom_attr(collector_name)

    def add_folder(self, collector_name:str, folder:str):
        '''
        Add a folder to the DataCollection object.

        Parameters
        ----------
        collector_name : str
            The name of the collector.
        folder : str
            The path of the folder.

        Returns
        -------
        None.

        '''

        self.collector[collector_name]['folder'] = folder

    def add_symbol_list(self, collector_name:str, symbol_list:list=[]): # optional
        '''
        Add a symbol list to the DataCollection object.
        
        Parameters
        ----------
        collector_name : str
            The name of the collector.
        symbol_list : list
            The symbol list.

        Returns
        -------
        None.

        '''

        self.collector[collector_name]['symbol_list'] = symbol_list

    def add_custom_attr(self, collector_name:str, **kwargs):
        '''
        Add custom attributes to the DataCollection object.

        Parameters
        ----------
        collector_name : str
            The name of the collector.
        kwargs : dict
            The custom attributes.

        Returns
        -------
        None.

        '''
        
        self.collector[collector_name]['custom'] = kwargs

    def adquire(self, id_integer=True, serializer='json', zip_extension=None, **kwargs):
        '''
        Adquire metadata from the collectors.

        Parameters
        ----------
        id_integer : bool
            A key feature is dealing with scenario results. The dimension 'id' represents the id of the scenarios. Each individual scenario id could be either numbered strings or integers.
        serializer : str
            Scenario metadata is provided through serialization formats. The default is 'json'. A JSON file has to be present in every scenario folder.
        zip_extension : str
            Scenario folders that contain files with optimization problem output can be contained as compressed files. The default is None. Options are 'zip' or '7z'.
        kwargs : dict
            The custom attributes.

        Returns
        -------
        None.

        '''

        self.scenarios_path = []
        self.config = {}
        self.data = []

        for collector in self.collector:
            folder = self.collector[collector]['folder']
            symbol_list = self.collector[collector]['symbol_list']
            custom_attr = self.collector[collector]['custom']
            parser = self.collector[collector]['parser']
            scenarios_path = _get_files_path_list(folder=folder, zip_extension=zip_extension, file_extension=serializer)
            self.scenarios_path += select_scenario_files(scenarios_path, zip_extension=zip_extension)
            for symbol_info_dict in parser(folder=folder, symbol_names=symbol_list, zip_extension=zip_extension, **kwargs):
                symbol_info_dict['collector'] = collector
                for attr in custom_attr:
                    symbol_info_dict[attr] = custom_attr[attr]
                self.data.append(symbol_info_dict)

        assert len(self.scenarios_path) > 0, "No scenario folder found"
        for scenario_path in self.scenarios_path:
            config = load_scenario_info(scenario_path,serializer=serializer, zip_extension=zip_extension)
            self.config[config['name']] = config
        self.config = dict(sorted(self.config.items()))
        assert len(self.config) == len(self.scenarios_path), "Config files with same id found"

        self._get_symbol_lists()
        self._scenario_name_shortener(id_integer)
        self._get_metadata_template()
        self._get_all_scenario_metadata()
        self._join_all_symbols()

    def _get_symbol_lists(self):
        list_of_symbols = []
        symbols_and_value_type = {}
        for symb_info in self.data:
            list_of_symbols.append(symb_info['symbol_name'])
            symbols_and_value_type[(symb_info['symbol_name'], symb_info['value_type'])] = None

        self.symbol_name_list = sorted(list(set(list_of_symbols)))
        self.symbol_valuetype_dict = dict(sorted(symbols_and_value_type.items()))

        return None
    
    def _scenario_name_shortener(self, id_integer=True):
        flag = False
        pattern = re.compile(r"(\d+)", re.IGNORECASE)
        names = []
        numbs = []
        shortnames = {}
        for scen in self.config:
            name = scen
            names.append(name)
            if pattern.search(name) != None:
                numbs.append(int(pattern.search(name)[0]))
            else:
                flag = True
        nrmax = max(numbs)
        for i in range(1,11):
            result = nrmax//10**i
            if result <= 1:
                digitM = i
                break

        names = sorted(names)
        number = len(names)
        for i in range(1,11):
            result = number//10**i
            if result <= 1:
                digitL = i
                break
        digit = max([digitL, digitM])
        if not flag:
            names_set = list(set(names))
            if len(names) == len(names_set):
                if len(names) == len(set(numbs)):
                    for name in names:
                        if id_integer:
                            shortname = int(pattern.search(name)[0])
                        else:
                            shortname = "S" + pattern.search(name)[0].zfill(digit)
                        shortnames[name] = shortname
                else:
                    flag = True
            else:
                flag = True
        if flag:
            for n, name in enumerate(names):
                if id_integer:
                    shortname = n
                else:
                    shortname = "S" + str(n).zfill(digit)
                shortnames[name] = shortname
        self.short_names = shortnames
        return None

    def _get_metadata_template(self):
        items_collector = []
        for scen in self.config:
            items_collector += list(self.config[scen]["metadata"])
        items = list(set(items_collector))
        self.metadata_template = {item:None for item in items}
        return None

    def _get_scenario_metadata(self, scen:str):
        scenario_metadata = {}
        for key in self.metadata_template:
            if key in self.config[scen]["metadata"]:
                scenario_metadata[key] = self.config[scen]["metadata"][key]
            else:
                scenario_metadata[key] = None
        return scenario_metadata

    def _get_all_scenario_metadata(self):
        all_scenario_metadata = {}
        for scen in self.config:
            all_scenario_metadata[scen] = self._get_scenario_metadata(scen)
        self.scenarios_metadata = all_scenario_metadata
        return None

    def _get_symbol_metadata(self, symbol_name:str, value_type:str):
        symbol_metadata = {}
        for scen in self.config:
            if (symbol_name,value_type) in self.symbol_valuetype_dict:
                symbol_metadata[scen] = self.scenarios_metadata[scen]
            else:
                print(f"{symbol_name} not found in {scen}")
                symbol_metadata[scen] = self.metadata_template
        return symbol_metadata

    def _join_scenarios_by_symbol(self, symbol_name:str, value_type:str='v'):
        """
        symbol
        """
        for data in self.data:
            if data['symbol_name'] == symbol_name and data['value_type'] == value_type:
                if self.symbols_book is None:
                    self.symbols_book = {}
                if (symbol_name, value_type) not in self.symbols_book:
                    self.symbols_book[(symbol_name, value_type)] = {}
                if 'short_names' not in self.symbols_book[(symbol_name, value_type)]:
                    self.symbols_book[(symbol_name, value_type)]['short_names'] = self.short_names
                if 'metadata' not in self.symbols_book[(symbol_name, value_type)]:
                    self.symbols_book[(symbol_name, value_type)]['metadata'] = self._get_metadata(self._get_symbol_metadata(symbol_name, value_type))
                if 'scenario_data' not in self.symbols_book[(symbol_name, value_type)]:
                    self.symbols_book[(symbol_name, value_type)]['scenario_data'] = {}
                self.symbols_book[(symbol_name, value_type)]['scenario_data'][data['scenario_name']] = data
        self.symbols_book[(symbol_name, value_type)]['scenario_data'] = dict(sorted(self.symbols_book[(symbol_name, value_type)]['scenario_data'].items()))

    def _get_metadata(self, raw_metadata):
        short_id = self.short_names
        dc = {}
        for k, v in raw_metadata.items():
            dc[short_id[k]] = {}
            for key, value in v.items():
                dc[short_id[k]][key] = value
        # new
        tdc = {}
        for ids, details in dc.items():
            for key in details:
                if key not in tdc:
                    tdc[key] = {}
                tdc[key][ids] = details[key]
        return tdc
        # return pd.DataFrame(dc).transpose().to_dict()

    def _join_all_symbols(self):
        for symb in self.symbol_valuetype_dict:
            self._join_scenarios_by_symbol(*symb)
        return None

    def __repr__(self):
        return '''DataCollection()'''





allowed_string = '.-_0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

value_type_name_map = {
                        "v": "Value",
                        "m": "Marginal",
                        "up": "Upper",
                        "lo": "Lower",
                        }


class Settings:
    def __init__(self):
        self.name_value_type_token = {}

    def append(self, name_value_type_token:tuple):
        assert isinstance(name_value_type_token, tuple), "A tuple of symbol name, value_type, and symbol_handler_token must be provided."
        assert name_value_type_token not in self.name_value_type_token, f"Symbol {name_value_type_token} already exists in a SymbolsHandler instance. Can not be loaded again."
        self.name_value_type_token[name_value_type_token] = True

    def exists(self, name_value_type_token:tuple):
        assert isinstance(name_value_type_token, tuple), "Symbol name, value_type, and symbol_handler_token must be provided."
        if name_value_type_token in self.name_value_type_token:
            return True
        else:
            return False

    def getnames(self):
        return self.name_value_type_token

settings = Settings()


# TODO: create a union of all array.coords used sofar. collect all uniques from coo from used arrays.
#       Then compare both to warn which element never used to improve efficiency


import os
import glob
import uuid
import itertools
import numpy as np
import karray as ka
import json
from json import JSONDecodeError
from typing import Union, List, Dict
from .handler import DataCollection
from .settings import settings, allowed_string, value_type_name_map


class SymbolsHandler:
    def __init__(self, method:str, folder_path:str=None, obj:DataCollection=None):
        ''' 
        Initialize SymbolsHandler object.

        Parameters
        ----------
        method : str
            The method used to collect raw variables or symbols. Two methods are available: 1. 'folder' - the symbols are collected from the folder specified by folder_path. 2. 'object' - the variables are collected from the object specified by obj which must be a DataCollection object.
        folder_path : str
            The path to the folder containing the symbols as feather files. Only used if method is 'folder'.
        obj : DataCollection
            The DataCollection object. Only used if method is 'object'.

        Returns
        -------
        None.

        Examples
        --------
        Build symbols from raw scenarios output (parser and loader from arrow tables and feather file format).

            >>> import symbolx as syx
            >>> from symbolx import DataCollection, SymbolsHandler
            >>> DC = DataCollection()
            >>> DC.add_collector(collector_name='opt_model', parser=syx.symbol_parser_feather, loader=syx.load_feather)
            >>> DC.add_folder(collector_name='opt_model', './raw_model_output')
            >>> DC.add_custom_attr(collector_name='opt_model', with_='pandas')
            >>> DC.adquire(id_integer=True, zip_extension=None)
            >>> SH = SymbolsHandler(method='object', obj=DC)

            
        Load symbols from folder.

            >>> from symbolx import , SymbolsHandler
            >>> SH = SymbolsHandler(method='folder', folder_path='./symbols_files')

            
        '''

        assert isinstance(method, str), "Arg 'method' must be a string."
        assert method in ["folder", "object"], "Arg 'method' must be either 'folder' or 'object'"
        self.method = method
        self.folder_path = None
        self.symbols_book = None
        self.symbol_handler_token = str(uuid.uuid4()) # TODO: this can be changed by hashing the input file
        self.saved_symbols = {}
        self.token_info = None
        self.input_method(method=method, folder_path=folder_path, obj=obj)

    def input_method(self, method:str, **kwargs):
        '''
        Input the method used to collect raw variables or symbols.

        Parameters
        ----------
        method : str
            The method used to collect raw variables or symbols. Two methods are available:
            1. 'folder' - the symbols are collected from the folder specified by folder_path.
            2. 'object' - the variables are collected from the object specified by obj which must be a DataCollection object.
        kwargs : dict
            Two keyword arguments are available:
            1. 'folder_path' - the path to the folder containing the symbols as feather files. Only used if method is 'folder'.
            2. 'obj' - the DataCollection object. Only used if method is 'object'.
        
        Returns
        -------
        None.

        '''

        if method == "object":
            self.from_object(obj=kwargs['obj'])
        elif method == "folder":
            self.from_folder(folder_path=kwargs['folder_path'])
        else:
            raise Exception('A method mus be provided from either "object" or "folder"')

    def from_object(self, obj:DataCollection):
        '''
        Extracts information from the DataCollection object, such as variable name and value type. Value type can be either 'value', 'marginal', 'lower', or 'upper' as define in GAMS for variables.

        It also populates a dictionary with the variable name, value type and symbol_handler_token—A symbol_handler_token is generated per instance; this helps to avoid mixup variables of other problems.

        Parameters
        ----------
        obj : DataCollection
            The DataCollection object.

        Returns
        -------
        None.

        '''

        self.symbols_book = obj.symbols_book
        self.collector = obj.collector
        self.short_names = obj.short_names
        for n_v in self.symbols_book:
            settings.append((*n_v,self.symbol_handler_token))


    def from_folder(self, folder_path:str=None):
        '''
        Extracts information from the folder, such as variable name and value type of all symbols saved with extension 'feather'.
        
        Parameters 
        ----------

        folder_path : str
            The path to the folder containing the symbols as feather files. Only used if method is 'folder'.

        Returns
        -------
        None.

        '''

        self.token_info = {}
        self.folder_path = folder_path
        files = glob.glob(os.path.join(self.folder_path, "*.feather"))
        self.symbols_book = {}
        for file in files:
            loaded_file_dict = from_feather_info(file)
            self.symbols_book[(loaded_file_dict['name'],loaded_file_dict['value_type'])] = file

            token = loaded_file_dict['symbol_handler_token']
            settings.append((loaded_file_dict['name'],loaded_file_dict['value_type'],loaded_file_dict['symbol_handler_token']))

            if token not in self.token_info:
                self.token_info[token] = {}
            self.token_info[token][(loaded_file_dict['name'],loaded_file_dict['value_type'])] = file
        if len(self.token_info) > 1:
            print(f"There are Symbols' files with different symbol_handler_token in {folder_path}")
            print("       Symbol's scenario short names or id's might be in conflict.")
            print("       Make sure all Symbols are from the same symbol_handler_token")
            print("       Otherwise, it may happen, for example, different scenarios have the same id.")
            print("       See 'token_info' attribute of SymbolsHandler for more information")


    def append(self, **kwargs):
        '''
        Store a symbol in a temporal container to posterior saving. The container can be accessed by the 'saved_symbols' attribute.

        Parameters
        ----------
        kwargs : dict
            The keyword arguments represents the name of the symbol. value argument is the symbol object.
        
        Returns
        -------
        None.
        
        '''

        for name in kwargs:
            symbol = kwargs[name]
            assert len(set(name).difference(set(allowed_string))) == 0, f"Symbol name '{name}' contains special characters. Please, change the name of the symbol. Allowed chars are: {allowed_string}"
            symbol.name = name
            self.saved_symbols[(name, symbol.value_type)] = symbol

    def save(self, folder_path=None, appended_or_all:str='all', gls:dict=None, overwrite:bool=False):
        '''
        Save the symbols in the temporary container to the folder specified by folder_path.

        Parameters
        ----------
        folder_path : str
            The path of the folder to save the symbols.
        appended_or_all : str
            If appended_or_all is 'all', all symbols contained in gls are saved. If appended_or_all is 'appended', only symbols contained in saved_symbols attribute are saved.
        gls : dictionary
            Recomended globals() that provides a dictionary with all variables from the script or notebook. The current method identifies the variables that are Symbols to be saved.
        overwrite : bool
            If overwrite is True, the symbols present in the folder are overwritten. If overwrite is False, the symbols in the folder are not overwritten. Default is False.

        Returns
        -------
        None.

        '''

        files = glob.glob(os.path.join(folder_path,'*.feather'))
        existing_pairs = {}
        print("Checking folder for duplicate symbol names...")
        for ff in files:
            file = os.path.basename(ff)
            name_sections = file.split('.')
            name = name_sections[0]
            value_type = name_sections[1]
            token = name_sections[2]
            assert (name,value_type) not in existing_pairs, f"Only one symbol name {name}.{value_type} should be found in '{folder_path}' folder. Two or more found\nTokens: {existing_pairs[(name,value_type)]} and {token}"
            existing_pairs[(name,value_type)] = token
        print("Checking passed")
        if folder_path is None:
            assert self.folder_path is not None, "folder_path must be provided"
            folder_path = self.folder_path
        os.makedirs(folder_path, exist_ok=True)
        if appended_or_all == 'all':
            symbol_map = []
            assert gls is not None, "if 'appended_or_all' is 'all' then provide gls=globals() to able to capture all variables that are Symbol instances"
            for name, symbol in gls.items():
                if isinstance(symbol, Symbol):
                    symbol.name = name
                    symbol_map.append(symbol)
        elif appended_or_all == 'appended':
            symbol_map = list(self.saved_symbols.values())
        else:
            raise Exception(f"Function argument 'appended_or_all' must be 'appended' or 'all'. Given: {appended_or_all}")

        for symbol in symbol_map:
            if not overwrite:
                assert (symbol.name,symbol.value_type) not in existing_pairs, f"Only one symbol name {symbol.name}.{symbol.value_type} should be saved in '{folder_path}' folder. Existing file found:\n{symbol.name}.{symbol.value_type}.{existing_pairs[(symbol.name,symbol.value_type)]}.feather"
            file_name = f"{symbol.name}.{symbol.value_type}.{symbol.symbol_handler_token}.feather"
            file_path = os.path.join(folder_path,file_name)
            symbol.to_feather(file_path)

    def get_info(self, symbol_name, value_type):
        '''
        Get information about a symbol.

        Parameters
        ----------
        symbol_name : str
            The name of the symbol.
        value_type : str
            The value type of the symbol.

        Returns
        -------
        symbol_info : dict or str
            if returns a dict, it contains metadata about the symbol and its scenarios. Only works if 'object' was the method used to create the symbol_handler.
            if returns a str, it contains the path to the symbol.feather file. Only works if 'folder' was the method used to create the symbol_handler.

        '''

        assert (symbol_name, value_type) in self.symbols_book, f"Pairs {(symbol_name, value_type)} not present in the symbols_book attribute of SymbolsHandler"
        if isinstance(self.symbols_book[(symbol_name, value_type)], dict):
            return self.symbols_book[(symbol_name, value_type)]
        elif isinstance(self.symbols_book[(symbol_name, value_type)], str):
            return self.symbols_book[(symbol_name, value_type)]

    def __repr__(self):
        return f'''SymbolsHandler(method='{self.method}')'''

def from_feather_dict(path, use_threads=True, with_="polars"):
    '''
    Load a feather file that contains a Symbol and returns a dictionary with symbol array and its metadata.

    Parameters
    ----------
    path : str
        The path to the feather file.
    use_threads : bool
        If True, the feather file is loaded in parallel. If False, the feather file is loaded in serial. Default is True.
    with_ : str
        The name of the python package used to load the arrow table from the feather file. Two options are available: 'polars' or 'pandas'. Default is 'polars'.
    
    Returns
    -------
    symbols_dict : dict
        A dictionary with symbol array and its metadata.

    '''

    import pyarrow.feather as ft
    arr = ka.from_feather(path, use_threads=use_threads, with_=with_)
    restored_table = ft.read_table(path, use_threads=use_threads)
    custom_meta_key = 'symbolx'
    restored_meta_json = restored_table.schema.metadata[custom_meta_key.encode()]
    restored_meta = recurse(restored_meta_json)
    return dict(array=arr,**restored_meta)

def from_feather_info(path):
    '''
    Load a feather file that contains a Symbol and returns a dictionary with symbol name, value type and symbol handler token.

    Parameters
    ----------
    path : str
        The path to the feather file.

    Returns
    -------
    symbols_info : dict
        A dictionary with symbol name, value type and symbol handler token.
    '''

    import pyarrow.feather as ft
    restored_table = ft.read_table(path)
    custom_meta_key = 'symbolx'
    restored_meta_json = restored_table.schema.metadata[custom_meta_key.encode()]
    restored_meta = json.loads(restored_meta_json)
    return dict(name=restored_meta['name'], value_type=restored_meta['value_type'], symbol_handler_token=restored_meta['symbol_handler_token'])

def recurse(d):
    if isinstance(d, dict):
        loaded_d = d
    elif isinstance(d, bytes):
        loaded_d = json.loads(d)
    else:
        return d
    new_dc = {}
    for k, v in loaded_d.items():
        if k.isdigit():
            r = int(k)
        else:
            r = k
        new_dc[r] = recurse(v)
    return new_dc

def build_array(symbol_name:str, value_type:str, symbol_handler:SymbolsHandler):
    """
    Build a symbol array collecting all variable arrays through all scenarios. Only works if 'object' was the method used to create the symbol_handler.
    
    Parameters
    ----------
    symbol_name : str
        The name of the symbol.
    value_type : str
        The value type of the symbol.
    symbol_handler : SymbolsHandler
        The symbol handler.

    Returns
    -------
    symbol_array : karray.Array
        The symbol array.

    """

    list_of_arrays = []
    for scenario_id in symbol_handler.get_info(symbol_name, value_type)['scenario_data']:
        array_with_id = insert_id_dim(symbol_name, value_type, scenario_id, symbol_handler)
        list_of_arrays.append(array_with_id)
    return ka.concat(list_of_arrays)

def insert_id_dim(symbol_name:str, value_type:str, scenario_id:str, symbol_handler:SymbolsHandler):
    """
    Creates an karray.Array from loader function. Then insert to the array a new dimension 'id' with the corresponding scenario_id.

    Parameters
    ----------
    symbol_name : str
        The name of the symbol.
    value_type : str
        The value type of the symbol.
    scenario_id : str
        The id of the scenario.
    symbol_handler : SymbolsHandler
        The symbol handler.

    Returns
    -------
    symbol_array : karray.Array
        The symbol array.

    """

    single_symbol = symbol_handler.get_info(symbol_name, value_type)['scenario_data'][scenario_id]
    single_array_dict = symbol_handler.collector[single_symbol['collector']]['loader'](**single_symbol)
    oarray = ka.Array(**single_array_dict)
    narray = oarray.add_dim(id=symbol_handler.short_names[scenario_id])
    return narray


class Symbol:
    def __init__(
        self,
        name: str=                      None,
        value_type: str=                'v',
        metadata: dict=                 None,
        array: ka.Array=                None,
        symbol_handler_token: str=      None,
        symbol_handler: SymbolsHandler= None,
        ):
        '''
        Symbol constructor.

        Parameters
        ----------
        name : str
            The name of the symbol.
        value_type : str
            The value type of the symbol.
        metadata : dict (optional)
            Nested dictionary with metadata about the symbol. The structure is as follows: {'metadata_1': {1: 'metadata_1_value of scenario 1', 2: 'metadata_2_value of scenario 2'}, ...}
        array : karray.Array
            The symbol array. This must be provided if 'symbol_handler' is not provided.
        symbol_handler_token : str
            The symbol handler token. This must be provided if 'symbol_handler' is not provided.
        symbol_handler : SymbolsHandler
            The symbol handler instance that contains information about the variables and scenarios or the symbol file path.

        Returns
        -------
        symbol : Symbol
            The symbol.

        Examples
        --------
        Based on a symbol_handler

            >>> var1 = Symbol(name='VAR1', symbol_handler=SH)
            >>> var2 = Symbol(name='VAR2', value_type='v', symbol_handler=SH)
            >>> var3 = Symbol(name='VAR3', value_type='m', symbol_handler=SH)

            
        Based on another symbol

            >>> var3_copy = Symbol(name='VAR3_copy', value_type='v', metadata=var3.metadata, array=var3.array, symbol_handler_token=var3.symbol_handler_token)})

            
        Based on a symbol file

            >>> var4 = syx.from_feather(symbol_file_path)

            
        '''
        self.__dict__["_repo"] = {}
        if symbol_handler is not None:
            if 'object' == symbol_handler.method:
                self.build(symbol_handler, name, value_type)
            elif 'folder' == symbol_handler.method:
                self.load(symbol_handler, name, value_type)
        else:
            if settings.exists((name,value_type,symbol_handler_token)):
                print("The symbol name has already been taken by another Symbol stored in SymbolsHandler.")
                print("Please choose another name to avoid overwriting the existing symbol when saving the current symbol.")
            self.name=                 name
            self.value_type=           value_type
            self.metadata=             metadata
            self.array=                array
            self.symbol_handler_token= symbol_handler_token
        self.check_input()
        # optional attributes
        self.dataframe = None


    def __setattr__(self, name, value):
        self._repo[name] = value

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(name) # ipython requirement for repr_html
        if name == "dataframe":
            if name in self._repo:
                if self._repo[name] is None:
                    self._repo[name] = self.array.to_pandas()
                    return self._repo[name].copy()
                else:
                    #Sanity check
                    df = self._repo[name]
                    dense = df.value.values.reshape(self.array.shape)
                    if np.allclose(dense, self.array.dense):
                        return df.copy()
                    else:
                        print(f"Getting dataframe again...")
                        self._repo[name] = self.array.to_pandas()
                        return self._repo[name].copy()
            else:
                self.definition_msg(name) # dev msg
        else:
            return self._repo[name]

    def build(self, symbol_handler:SymbolsHandler, name: str=None, value_type: str=None):
        '''
        Populate the symbol instance with information from the symbol handler and the loader function provided to the DataCollection instance.

        Parameters
        ----------
        symbol_handler : SymbolsHandler
            The symbol handler instance that contains information about the variables and scenarios. This method is used when the symbol handler is instantiated with 'object' method.
        name : str
            The name of the symbol.
        value_type : str
            The value type of the symbol.

        Returns
        -------
        symbol : Symbol
            The symbol.

        '''

        assert isinstance(symbol_handler, SymbolsHandler), "'symbol_handler' must be a SymbolsHandler object"
        assert isinstance(name, str), "'name' must be a string"
        assert isinstance(value_type, str), "'value_type' must be a string"
        print(f"{name:<25} value_type: {value_type_name_map[value_type]} ({value_type})")

        key = (name,value_type)
        assert key in symbol_handler.symbols_book, f"'{key}' is not present in 'symbol_handler.symbols_book' dictionary"

        self.name=                 name
        self.value_type=           value_type
        self.metadata=             symbol_handler.get_info(*key)['metadata']
        self.array=                build_array(name, value_type, symbol_handler)
        self.symbol_handler_token= symbol_handler.symbol_handler_token
        
    def load(self, symbol_handler:SymbolsHandler, name: str=None, value_type: str=None):
        '''
        Symbol is recreated from feather file. This method is used when the symbol handler is instantiated with 'folder' method providing the path of the symbol file.

        Parameters
        ----------
        symbol_handler : SymbolsHandler
            The symbol handler instance that contains information about the variables and scenarios. This method is used when the symbol handler is instantiated with 'folder' method.
        name : str
            The name of the symbol.
        value_type : str
            The value type of the symbol.

        Returns
        -------
        symbol : Symbol
            The symbol.

        '''

        assert isinstance(symbol_handler, SymbolsHandler), "'symbol_handler' must be a SymbolsHandler object"
        assert isinstance(name, str), "'name' must be a string"
        assert isinstance(value_type, str), "'value_type' must be a string"
        print(f"{name:<25} value_type: {value_type_name_map[value_type]} ({value_type})")

        key = (name,value_type)
        assert key in symbol_handler.symbols_book, f"'{key}' is not present in 'symbol_handler.symbols_book' dictionary"

        file_path = symbol_handler.get_info(*key)
        symbol_dict = from_feather_dict(file_path, with_=ka.settings.feather_with)
        self.name=                 name
        self.value_type=           value_type
        self.metadata=             symbol_dict['metadata']
        self.array=                symbol_dict['array']
        self.symbol_handler_token= symbol_dict['symbol_handler_token']

    def to_arrow(self):
        '''
        Convert the symbol to an arrow table with metadata. The symbol array contains karray array and the karray metadata. The returned table contains karray and symbox metadata as well.

        Returns
        -------
        arrow_table : pyarrow.Table
            The arrow table.

        '''

        table = self.array.to_arrow()
        existing_meta = table.schema.metadata
        custom_meta_key = 'symbolx'
        custom_metadata = {}
        attr = ['name', 'value_type', 'metadata','symbol_handler_token']
        for k,v in self._repo.items():
            if k in attr:
                custom_metadata[k] = v

        custom_meta_json = json.dumps(custom_metadata)
        existing_meta = table.schema.metadata
        combined_meta = {custom_meta_key.encode() : custom_meta_json.encode(),**existing_meta}
        table = table.replace_schema_metadata(combined_meta)
        return table

    def to_feather(self, path:str):
        '''
        Convert the symbol to a feather file with metadata. The symbol array contains karray array and the karray metadata. The returned table contains karray and symbox metadata as well.

        Parameters
        ----------
        path : str
            The path of the feather file.

        Returns
        -------
        None

        '''

        import pyarrow.feather as ft
        sets = set(allowed_string)
        sets.add(os.path.sep)
        joint = ''.join(sorted(sets))
        assert len(set(path).difference(sets)) == 0, f"There are/is special characters in path '{path}'. Allowed chars are: {joint}"

        table = self.to_arrow()
        ft.write_feather(table, path)
        print(f"{path}")
        return None

    def check_input(self):
        '''
        Check the input of the symbol. 

        Returns
        -------
        None

        '''

        assert self.name is not None and isinstance(self.name, str), "Name of symbol must be provided."
        assert self.value_type is not None and isinstance(self.value_type, str), "Value type of symbol must be provided."
        assert self.metadata is not None and isinstance(self.metadata, dict), "metadata of symbol must be provided."
        assert self.array is not None and isinstance(self.array, ka.Array), "Array must be provided."
        assert self.symbol_handler_token is not None and isinstance(self.symbol_handler_token, str), "Symbol handler token must be provided."

    @property
    def dims(self):
        '''
        Get the dimensions of the symbol.

        Returns
        -------
        dims : list of strings
            The dimensions of the symbol.

        '''
        return self.array.dims

    @property
    def df(self):
        '''
        Get the dataframe of the symbol with the dimenssions coordinates as multiindex. It contains all possible combinations of the dimensions coordinates.

        Returns
        -------
        df : pandas.DataFrame
            The multiindex dataframe of the symbol. Long data format.

        '''

        return self.dataframe.set_index(self.array.dims)

    @property
    def dfm(self):
        '''
        Get the dataframe of the symbol with metadata as additional columns. Dimenssions coordinates are as columns.

        Returns
        -------
        dfm : pandas.DataFrame
            The dataframe of the symbol with long data format.

        '''

        dfm = self.dataframe
        for k, v in self.metadata.items():
            dfm[k] = dfm["id"].map(v)
        return dfm

    @property
    def dfc(self):
        '''
        Get the dataframe of the symbol with only custom metadata as additional columns. Dimenssions coordinates are as columns.

        Returns
        -------
        dfc : pandas.DataFrame
            The dataframe of the symbol with long data format.

        '''

        dfc = self.dataframe
        for k, v in self.metadata.items():
            if 'custom_' in k:
                dfc[k] = dfc["id"].map(v)
        return dfc
    
    def to_polars(self):
        '''
        Convert the symbol to a polar data frame.

        Returns
        -------
        polar_df : polars.DataFrame
            The polar data frame of the symbol.

        '''

        return self.array.to_polars()

    def to_pandas(self):
        '''
        Convert the symbol to a pandas data frame.

        Returns
        -------
        pandas_df : pandas.DataFrame
            The pandas data frame of the symbol.

        '''

        return self.array.to_pandas()

    def metadata_union(self, other=None):
        '''
        Union the metadata of two symbols.

        Parameters
        ----------
        other : Symbol

        Returns
        -------
        metadata : dict
            The union of the metadata of two symbols.

        '''

        self_metadata = self.metadata
        other_metadata = other.metadata
        new_metadata = {}
        for elem in self_metadata.keys():
            new_metadata[elem] = {**self_metadata[elem],**other_metadata[elem]}
        return new_metadata

    def new_symbol(self, array, new_name, other=None):
        '''
        Create a new symbol with a new name and a new array.

        Parameters
        ----------
        array : ka.Array
            The array of the new symbol.
        new_name : str
            The name of the new symbol.
        other : Symbol
            The other symbol.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        if isinstance(other, Symbol):
            assert self.symbol_handler_token == other.symbol_handler_token, "Symbol handler tokens must be the same"
            new_metadata = self.metadata_union(other)
        elif other is None or isinstance(other,(int,float)):
            new_metadata = self.metadata
        else:
            raise Exception(f'other must be either a Symbol object, int, float or None, but it is: {str(type(other))}')
        new_object = Symbol(name=new_name, value_type='v',
                            metadata=new_metadata, array=array, 
                            symbol_handler_token=self.symbol_handler_token)
        return new_object

    def __add__(self, other):
        '''
        Add two symbols or a symbol and an number.

        Parameters
        ----------
        other : Symbol or int or float
            The symbol or number to add.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        if isinstance(other, (int, float)):
            new_array = self.array + other
            new_name =  f"({self.name})+{str(other)}"
            return self.new_symbol(new_array, new_name, other)
        elif isinstance(other, Symbol):
            new_array = self.array + other.array
            new_name = f"({self.name})+({other.name})"
            return self.new_symbol(new_array, new_name, other)
        else:
            raise Exception(f'{type(other)} is not supported')

    def __sub__(self, other):
        '''
        Subtract two symbols or a symbol and an number.

        Parameters
        ----------
        other : Symbol or int or float
            The symbol or number to subtract.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        if isinstance(other, (int, float)):
            new_array = self.array - other
            new_name =  f"({self.name})-{str(other)}"
            return self.new_symbol(new_array, new_name, other)
        elif isinstance(other, Symbol):
            new_array = self.array - other.array
            new_name = f"({self.name})-({other.name})"
            return self.new_symbol(new_array, new_name, other)
        else:
            raise Exception(f'{type(other)} is not supported')

    def __mul__(self, other):
        '''
        Multiply two symbols or a symbol and an number.

        Parameters
        ----------
        other : Symbol or int or float
            The symbol or number to multiply.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        Exceptions
        ----------
        Exception 1 : if the two symbols have no common dimensions.
        Exception 2 : if the two symbols have more than one uncommon dimensions.
        Exception 3 : if the class of other is not a Symbol or int or float.

        '''

        if isinstance(other, (int, float)):
            new_array = self.array * other
            new_name = f"({self.name})*{str(other)}"
            return self.new_symbol(new_array, new_name, other)
        elif isinstance(other, Symbol):
            diffdims = list(set(self.dims).symmetric_difference(other.dims))
            lendiff = len(diffdims)
            if set(self.dims) == set(other.dims):
                new_array = self.array * other.array
                new_name = f"({self.name})*({other.name})"
                return self.new_symbol(new_array, new_name, other)
            elif lendiff == 1:
                new_array = self.array * other.array
                new_name = f"({self.name})*({other.name})"
                return self.new_symbol(new_array, new_name, other)
            elif lendiff > 1:
                common_dims = list(set(self.dims).intersection(other.dims))
                if len(common_dims) > 0:
                    new_array = self.array * other.array
                    new_name = f"({self.name})*({other.name})"
                    return self.new_symbol(new_array, new_name, other)
                else:
                    raise Exception(f"The difference in dimensions is greater than one: '{diffdims}' and has no common dimensions")
        else:
            raise Exception(f'{type(other)} is not supported')

    def __truediv__(self, other):
        '''
        Divide two symbols or a symbol and an number.
        
        Parameters
        ----------
        other : Symbol or int or float
            The symbol or number to divide.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        Exceptions
        ----------
        Exception 1 : if the difference in dimensions is greater than one.
        Exception 2 : if the class of other is not a Symbol or int or float.

        '''

        if isinstance(other, (int, float)):
            new_array = self.array / other
            new_name = f"({self.name})/{str(other)}"
            return self.new_symbol(new_array, new_name, other)
        elif isinstance(other, object):
            diffdims = set(self.dims).symmetric_difference(other.dims)
            lendiff = len(diffdims)
            if set(self.dims) == set(other.dims):
                new_array = self.array / other.array
                new_name = f"({self.name})/({other.name})"
                return self.new_symbol(new_array, new_name, other)
            elif lendiff == 1:
                new_array = self.array / other.array
                new_name = f"({self.name})/({other.name})"
                return self.new_symbol(new_array, new_name, other)

            elif lendiff > 1:
                raise Exception(f"The difference in dimensions is greater than one: '{diffdims}'")
        else:
            raise Exception("The second term is not known, must be a int, float or a Symbol object")

    def __radd__(self, other):
        if isinstance(other, (int, float)):
            new_array =  other + self.array
            new_name = f"{str(other)}+({self.name})"
            return self.new_symbol(new_array, new_name, other)

    def __rsub__(self, other):
        if isinstance(other, (int, float)):
            new_array = other - self.array
            new_name = f"{str(other)}-({self.name})"
            return self.new_symbol(new_array, new_name, other)

    def __rmul__(self, other):
        if isinstance(other, (int, float)):
            new_array = other*self.array
            new_name = f"{str(other)}*({self.name})"
            return self.new_symbol(new_array, new_name, other)

    def __rtruediv__(self, other):
        if isinstance(other, (int, float)):
            new_array = other/self.array
            new_name = f"{str(other)}/({self.name})"
            return self.new_symbol(new_array, new_name, other)

    def rename(self, new_name: str):
        '''
        Rename the symbol.

        Parameters
        ----------
        new_name : str
            The new name of the symbol.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        return self.new_symbol(self.array, new_name)

    def dimreduc(self, dim:str='h', aggfunc=np.add.reduce):
        '''
        Reduce one dimension of the symbol by applying a numpy ufunc over all coordinates of the dimension.
        
        Parameters
        ----------
        dim : str
            The dimension to reduce.
        aggfunc : ufunc
            The numpy ufunc to apply. Options are [np.add.reduce,np.multiply.reduce,np.average]. Default np.add.reduce

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        new_array = self.array.reduce(dim, aggfunc)
        new_name = f"({self.name}).dimreduc({dim},{aggfunc})"
        return self.new_symbol(new_array, new_name)

    @property
    def items(self):
        '''
        Get the items of the symbol or coordinates.

        Returns
        -------
        coords: dict
            The items of the symbol or coordinates.

        '''

        return self.array.coords

    def rename_dim(self, **kwargs):
        """ 
        Rename a dimension of the symbol.

        Parameters
        ----------
        kwargs : dict
            The keyword arguments can be the current dimensions name and the value arguments are the new name of the dimensions.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        """

        new_array = self.array.rename(**kwargs)
        new_name = f"({self.name}).rename_dim(**{kwargs})"
        return self.new_symbol(new_array, new_name)
    
    def add_dim(self, dim_name: str, value: Union[str,int,dict]):
        '''
        Deprecated: Use add_dims instead.

        Add a dimension to the symbol.

        Parameters
        ----------
        dim_name : str
            The name of the dimension.
        value : Union[str,int,dict]
            if value is a string, the dimension column will contain this value only. if value is a dict, the dict must look like {column_header:{column_element: new_element_name}} where column_header must currently exists and all column_elements must have a new_element_name.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        if isinstance(value, (str,int)):
            new_array = self.array.add_dim(**{dim_name:value})
            new_name = f"({self.name}).add_dim({dim_name},{value})"
            return self.new_symbol(new_array, new_name)

        elif isinstance(value, dict):
            new_array = self.array.add_dim(**{dim_name:value})
            new_name = f"({self.name}).add_dim({dim_name},{value})"
            return self.new_symbol(new_array, new_name)
        else:
            raise Exception('value is neither str nor dict')
        
    def add_dims(self, **kwargs):
        '''
        Add a dimension to the symbol.

        Parameters
        ----------
        kwargs : dict
            The keyword arguments must be new dimension names and the value arguments contain information of the new coordinates as folows:
            value arguments : Union[str, int, dict[str, dict[str,str]], dict[str,list]]
            if value is a string or integer, the new dimension will contain this value only as a unique coordinate.
            if value is a dict of dict, the dict must look like {existing dimension mane:{current coordinate: new coordinate}}
            if value is a dict of list, the dict must look like {existing dimension mane:[numpy array of current coordinate, numpy array of new coordinate]} both arrays must have the same length. Numpy array of current coordinate must be unique. The position of the elements of the arrays is considered for the replacement.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        new_array = self.array.add_dim(**kwargs)
        new_name = f"({self.name}).add_dims({','.join(list(kwargs))})"
        return self.new_symbol(new_array, new_name)

    def dropna(self):
        '''
        Drop the NaN values from the symbol.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        new_array = self.array.dropna()
        new_name = f"({self.name}).dropna()"
        return self.new_symbol(new_array, new_name)

    def dropinf(self, pos:bool, neg:bool):
        '''
        Drop the inf values from the symbol.
        
        Parameters
        ----------
        pos : bool
            If True, drop the inf values from the positive direction.
        neg : bool
            If True, drop the inf values from the negative direction.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        new_array = self.array.dropinf(pos,neg)
        new_name = f"({self.name}).dropna({pos=},{neg=})"
        return self.new_symbol(new_array, new_name)

    def drop(self, dims:Union[str,List[str]]):
        '''
        Drop the specified dimensions from the symbol.

        Parameters
        ----------
        dims : Union[str,List[str]]
            The dimensions to drop.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        new_array = self.array.drop(dims=dims)
        new_name = f"({self.name}).dropna({dims=})"
        return self.new_symbol(new_array, new_name)

    def round(self, decimals:int):
        '''
        Round the symbol to the specified number of decimals.

        Parameters
        ----------
        decimals : int
            The number of decimals to round to.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        new_array = self.array.round(decimals=decimals)
        new_name = f"({self.name}).round({decimals=})"
        return self.new_symbol(new_array, new_name)

    def elems_to_datetime(self, new_dim:str, actual_dim:str, reference_date:str='01-01-2030', freq:str='H', sort_corrds:bool=True):
        '''
        Convert the elements of the symbol to datetime.

        Parameters
        ----------
        new_dim : str
            The new dimension name.
        actual_dim : str
            The actual dimension name.
        reference_date : str
            The reference date.
        freq : str
            The frequency of the datetime.
        sort_corrds : bool
            If True, sort the coordinates of the symbol.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        new_array = self.array.elems_to_datetime(new_dim=new_dim, actual_dim=actual_dim, reference_date=reference_date, freq=freq, sort_coords=sort_corrds)
        new_name = f"({self.name}).elems_to_datetime({new_dim}{actual_dim},{reference_date},{freq},{sort_corrds})"
        return self.new_symbol(new_array, new_name)

    def elems_to_int(self, new_dim:str, actual_dim:str):
        '''
        Convert the elements of the symbol to integers.

        Parameters
        ----------
        new_dim : str
            The new dimension name.
        actual_dim : str
            The actual dimension name.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        '''

        new_array = self.array.elems_to_int(new_dim, actual_dim)
        new_name = f"({self.name}).elems_to_int({new_dim=},{actual_dim=})"
        return self.new_symbol(new_array, new_name)

    def find_ids(self, **kwargs):
        '''
        Find scenarios ids in a symbol that comply with the criteria given by the keyword arguments. Several criteria can be specified where the result would be the intersection of all criteria.

        Parameters
        ----------
        kwargs : dict
            The keyword arguments must be the following:

            metadata name : str
                Metadata feature name.
            criteria : Tuple[str,Union[str,int,float]]
                The criteria to match. It is a tuple of two elements. The first element is a string that represents an operator such as '>', '<', '<=', '>=', '==', or '!='. The second element is the value to match.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        Examples
        --------
        >>> Z.find_ids(**{'time_series_scen':('==','NaN'),'co2price(n,tech)':('<',80)})

        '''

        dc = self.metadata
        collector = []
        for k,v in dc.items():
            if k in kwargs.keys():
                flag = False
                nan_str = False
                id_list = []
                for k2, v2 in v.items():
                    if isinstance(v2,str):
                        if isinstance(kwargs[k][1],str):
                            if eval(f"'{v2}' {kwargs[k][0]} '{kwargs[k][1]}'"):
                                id_list.append(k2)
                                if not flag:
                                    flag = True
                            if kwargs[k][1] != 'NaN' and v2 == 'NaN':
                                nan_str = True
                        else:
                            continue
                    elif np.isnan(v2):
                        if np.isnan(kwargs[k][1]):
                            if kwargs[k][0] == '==':
                                id_list.append(k2)
                                if not flag:
                                    flag = True
                        else:
                            if kwargs[k][0] == '!=':
                                id_list.append(k2)
                                if not flag:
                                    flag = True
                            elif kwargs[k][0] != '==':
                                print(f"{k} in {k2} has NaN value for condition '{kwargs[k][0]} {str(kwargs[k][1])}'. Not included")
                            
                    elif np.isnan(kwargs[k][1]):
                        if kwargs[k][0] == '!=':
                            id_list.append(k2)
                            if not flag:
                                flag = True
                        else:
                            continue
                    elif eval(f"{v2} {kwargs[k][0]} {kwargs[k][1]}"):
                        id_list.append(k2)
                        if not flag:
                            flag = True
                collector.append(set(id_list))
                if not flag:
                    print(f"Column '{k}' does not contain '{kwargs[k][1]}'")
                if nan_str:
                    print(f"Column '{k}' has 'NaN' as string. You can filter such string too.")
        not_present = []
        for cond in kwargs.keys():
            if cond in dc.keys():
                pass
            else:
                not_present.append(cond)
        if not_present:
            str_cond = ";".join(not_present)
            print(f"{str_cond} not in symbol's data")
        return set.intersection(*collector)

    def id_info(self, ID:Union[str,int]):
        '''
        Get information about a scenario id.

        Parameters
        ----------
        ID : Union[str,int]
            The scenario id.

        Returns
        -------
        dc : dict
            A dictionary with the following  as keys and the corresponding value.

        Examples
        --------

           >>> Z.id_info('S0001')

        '''

        dc = dict()
        for k, v in self.metadata.items():
            if ID in v.keys():
                dc[k] = v[ID]
        return dc
    
    def shrink(self, neg: Union[bool, list]=False, **kwargs):
        ''' 
        Shrinks the symbol to keep only those coordinates that comply the given criteria.
        karg is a dictionary of symbol sets as key and elements of the set as value.
        sets and elements must be present in the symbol.

        Parameters
        ----------
        neg : Union[bool, List[bool]]
            If True, the symbol is shrunk to keep only those coordinates that do not comply the criteria. If False, the symbol is shrunk to keep only those coordinates that comply the criteria.
            If a list of booleans, follows the same order as the dimensions provided as kwargs.
            Defaut is False.
        kwargs : dict
            The keyword arguments must be the following:

            dimension name : str (dictionary keys)
                dimesion name with the coordinates to keep if neg is False.
            coordinates list: list (dictionary values)
                    list of coordinates to keep if neg is True.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        Examples
        --------
        >>> Z.shrink(neg=[False, True], **{'tech':['pv','bio'],'h':[1,2]})

        >>> Z.shrink(neg=[False, True], tech=['pv','bio'], h=[1,2])

        '''

        if len(kwargs) > 1:
            if isinstance(neg,list):
                assert len(neg) == len(kwargs)
            else:
                new_neg = []
                for _ in range(len(kwargs)):
                    new_neg.append(neg)
                neg = new_neg
        else:
            neg = [neg]

        for key, value in kwargs.items():
            if key in self.dims:
                if set(value).issubset(self.items[key]):
                    pass
                else:
                    not_present = set(value) - (set(value) & set(self.items[key]))
                    present = (set(value) & set(self.items[key]))
                    if len(present) == 0:
                        raise Exception(f"{not_present} is/are not in {self.items[key]}")
                    else:
                        # print(f"    Only {present} exist in {self.items[key]} and meet criteria at '{key}', while not {not_present}")
                        kwargs[key] = sorted(present)
            else:
                raise Exception(f"'{key}' is not in {self.dims} for symbol {self.name}")

        right_kwargs = {k:v for k,v in kwargs.items()}
        i = 0
        for key in kwargs:
            if neg[i]:
                all_elems = self.items[key].tolist()
                for skip_elem in kwargs[key]:
                    all_elems.remove(skip_elem)
                right_kwargs[key] = all_elems

        new_array = self.array.shrink(**right_kwargs)
        new_name = f"({self.name}).shrink(neg={neg},{','.join(['='.join([k,str(v)]) for k,v in kwargs.items()])})"
        return self.new_symbol(new_array, new_name)

    def shrink_by_attr(self, neg=False, **kwargs):
        ''' 
        shrink_by_attributes generates new symbol based on metadata attributes of the symbol. Attributes can be seen with symbol_object.metadata.
        Shrink the symbol to keep only the coordinates that comply with the criteria given in kwargs.

        Parameters
        ----------
        neg : bool
            If True, the symbol is shrunk to keep only those ids that do not comply the criteria. If False, the symbol is shrunk to keep only those ids that comply the criteria.
            Defaut is False.
        kwargs : dict
            The keyword arguments must be the following:

            metadata name : str (dictionary keys)
                metadata name to consider as criteria.
            metadata elements : str (dictionary values)
                    metadata elements to keep if neg is False. If neg is True, the metadata elements that do not comply the criteria are kept.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        Examples
        --------
        >>> Z.shrink_by_attr(**{'run':[0,1], 'country_set':['NA']})

        >>> Z.shrink_by_attr(run=[0,1], country_set=['NA'])

        '''
        for key, value in kwargs.items():
            dc = self.metadata
            if key not in dc.keys():
                raise Exception(f"'{key}' is not in {list(dc.keys())} for symbol {self.name}")
        
        id_list = sorted(set([item for sublist in list(self.create_mix(kwargs).values()) for item in sublist]))
        new_object = self.shrink(id=id_list, neg=neg)
        new_object.name = f"(neg={neg},{self.name}).shrink_by_attr({','.join(['='.join([k,str(v)]) for k,v in kwargs.items()])})"
        return new_object

    def refdiff(self, reference_id:Union[int,str]=0):
        ''' 
        Returns the difference between the reference scenario (with its respective id) and the rest of the scenarios in a symbol. The symbol must have the dimension name 'id'.
        
        Parameters
        ----------
        reference_id : Union[int,str]
            The id of the reference scenario.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        Examples
        --------

        >>> Z.refdiff(reference_id='S0001')
        
        '''
        new_object = self - self.shrink(id=[reference_id]).dimreduc('id')
        new_object.name = f"({self.name}).refdiff({reference_id})"
        return new_object

    def create_mix(self, criteria):
        ''' 
        Private function used for refdiff_by_sections.
        '''

        combination = self.create_combination(criteria)
        order = criteria.keys()
        return self._find_ids_by_tuple(order,combination)

    def create_combination(self, criteria: dict):
        '''
        Private function used for refdiff_by_sections.
        '''

        return list(itertools.product(*criteria.values()))


    def _find_ids_by_tuple(self,key_order,combination):
        '''
        Private function used for refdiff_by_sections.
        '''

        groups = {}
        for i, pair in enumerate(combination):
            config = {}
            for k, v in zip(key_order, pair):
                config[k] = ('==',v)
            groups[i] = list(self.find_ids(**config))
        return groups

    def _ref_diff_group(self,refs,groups, verbose=False):
        '''
        Private function used for refdiff_by_sections.
        '''

        symbols = []
        for key in groups:
            if len(refs[key]) == 0:
                if verbose:
                    print(f"{refs} for key = {key} no reference id found")
                    print(groups)
                continue
            else:
                refdiff_symbol = self.shrink(id=list(groups[key])).refdiff(refs[key][0])
                symbols.append(refdiff_symbol)
        return sum(symbols)

    def refdiff_by_sections(self, criteria_dict, criteria_ref_dict, verbose=False):
        '''
        Returns the difference between the reference scenario following criteria based on the metadata attributes. 

        Parameters
        ----------
        criteria_dict : dict
            The criteria dictionary contains the metadata attributes as keys and the criteria as values.
        criteria_ref_dict : dict
            The reference criteria dictionary.
        verbose : bool
            If True, prints the results. Defaut is False.

        Returns
        -------
        new_symbol : Symbol
            The new symbol.

        Examples
        --------
        
        >>> Z.refdiff_by_sections(**{'run':[0,1], 'country_set':['NA']})

        '''

        groups = self.create_mix(criteria_dict)
        criteria_ref_full = {**criteria_dict,**criteria_ref_dict}
        refs = self.create_mix(criteria_ref_full)
        assert all([len(refs[key]) <= 1 for key in refs]), f"It should be only one reference scenario per group {criteria_ref_full.keys()} but more were found: {refs}"
        return self._ref_diff_group(refs,groups,verbose)

    def definition_msg(self, name):
        print(f"Attribute '{name}' must be defined first in __init__ method")

    def __repr__(self):
        return f'''Symbol(name='{self.name}', \n       value_type='{self.value_type}')'''


def from_feather(path):
    return Symbol(**from_feather_dict(path, with_=ka.settings.feather_with))


# symbolx
# copyright 2022, Carlos Gaete-Morales, DIW-Berlin 
"""
    symbolx
    copyright 2022, Carlos Gaete-Morales, DIW-Berlin 
"""
__version__ = (0, 4, 1)
__author__ = 'Carlos Gaete-Morales'


from .parsers.parser_csv import symbol_parser_csv, load_csv
from .parsers.parser_feather import symbol_parser_feather, load_feather
from .parsers.parser_gdx import symbol_parser_gdx, load_gdx
from .handler import DataCollection
from .utils import compress_subdirs, compress_dir, unzip_all, unzip
from .symbols import SymbolsHandler, Symbol, from_feather
from .settings import settings






import os
import zipfile
import tempfile
import numpy as np
import platform
import subprocess
from ..utils import _get_files_path_list, _convert_symbol_name_to_tuple


def symbol_parser_gdx(folder: str, symbol_names: list=[], zip_extension=None, **kwargs):
    '''
    Parse all symbols from a folder and returns a dictionary
    '''
    gams_dir = kwargs['gams_dir'] if 'gams_dir' in kwargs else None
    ret, gams_dir = get_gams_dir(gams_dir=gams_dir)
    if not ret:
        raise Exception("Gams directory not found")

    symbol_dict_with_value_type = {}
    for symbs in symbol_names:
        symb_tp = _convert_symbol_name_to_tuple(symbs)
        symbol_dict_with_value_type[symb_tp] = None

    file_list = _get_files_path_list(folder=folder, zip_extension=zip_extension, file_extension='gdx')

    symbol_list = []
    for file in file_list:
        scen_id = os.path.basename(file).split('.')[0]
        if zip_extension is not None:
            path_parts = file.split(zip_extension+os.sep)
            zip_fpath = path_parts[0] + zip_extension
            target_fpath = path_parts[1]
            with zipfile.ZipFile(zip_fpath, mode="r") as zip_io:
                listOfFileNames = zip_io.namelist()
                assert target_fpath in listOfFileNames
                with tempfile.TemporaryDirectory() as tmpdirname:
                    zip_io.extract(target_fpath, tmpdirname)
                    tempfile_path = os.path.join(tmpdirname,target_fpath)
                    symbol_items = _symbols_list_from_gdx(tempfile_path, gams_dir)
        else:
            symbol_items = _symbols_list_from_gdx(file, gams_dir)

        for (name, symb_type, nrdims) in symbol_items:
            if symb_type == 0: # set
                options = []
            elif symb_type == 1: # parameter
                options = ['v']
            elif symb_type == 2: # variable
                options = ['v', 'm']
            elif symb_type == 3: # equation
                options = ['v', 'm']
            for value_type in options:
                symb_tp = (name, value_type)
                if symb_tp in symbol_dict_with_value_type or len(symbol_names) == 0:
                                        # This fields are mandatory for a parser
                    symbol_list.append({'symbol_name':symb_tp[0],
                                        'value_type':symb_tp[1],
                                        'path':file,
                                        'scenario_name':scen_id,
                                        # Until here
                                        # you can add more (custom) attributes. It must be added also see handler.py def add_custom_attr() and be an attribute for loader
                                        'gams_dir':gams_dir,
                                        'zip_extension':zip_extension,
                                        'inf_to_zero':True, # included with default value. This can be changed later in handler.py def add_custom_attr()
                                        'verbose':False,
                                        })
    return symbol_list

def load_gdx(symbol_name: str, value_type: str='v', path: str='', gams_dir: str= None, inf_to_zero:bool=True, verbose:bool=False, zip_extension=None, **kwargs):
    '''
    Load custom GDX file.

    Parameters
    ----------
    symbol_name : str
        Name of the symbol to be extracted.
    value_type : str, optional
        Type of the symbol to be extracted. The default is 'v'.
    path : str
        Path to the gdx file.
    gams_dir : str, optional

    '''
    value_types = {'v':0, 'm':1, 'lo':2, 'up':3, 'scale':4}
    assert value_type in value_types.keys(), f'value_type must be one of the following: {value_types.keys()}'
    if zip_extension is not None:
        path_parts = path.split(zip_extension+os.sep)
        zip_fpath = path_parts[0] + zip_extension
        target_fpath = path_parts[1]
        with zipfile.ZipFile(zip_fpath, mode="r") as zip_io:
            listOfFileNames = zip_io.namelist()
            assert target_fpath in listOfFileNames
            with tempfile.TemporaryDirectory() as tmpdirname:
                zip_io.extract(target_fpath, tmpdirname)
                tempfile_path = os.path.join(tmpdirname,target_fpath)
                metadata = _gdx_get_symbol_data_dict(symbol_name=symbol_name, gdx_file=tempfile_path, gams_dir=gams_dir)
                symbol = _gdx_get_symbol_array_str(symbol_name=symbol_name, gdx_file=tempfile_path, gams_dir=gams_dir)
                nrdims = len(metadata['dims'])
                col_index = nrdims + value_types[value_type]
                raw_coo = symbol[:, list(range(nrdims)) + [col_index]]
                # Warning: gams2numpy pkg convert EPS to INF as 5e+300
                if inf_to_zero:
                    EPS = raw_coo[:, nrdims] == np.float64("5e+300")
                    raw_coo[EPS, nrdims] = 0.0
                    if verbose:
                        if sum(EPS) > 0:
                            print('GAMS EPS to 0.0 changed')
                value = raw_coo[:, nrdims].astype(float)
                index = {dim: raw_coo[:,idx] for idx, dim in enumerate(metadata['dims'])}
                coords = {dim: metadata['coords'][dim] for dim in metadata['dims']}
                return {'data': (index, value),'coords': coords}
    else:
        metadata = _gdx_get_symbol_data_dict(symbol_name=symbol_name, gdx_file=path, gams_dir=gams_dir)
        symbol = _gdx_get_symbol_array_str(symbol_name=symbol_name, gdx_file=path, gams_dir=gams_dir)
        nrdims = len(metadata['dims'])
        col_index = nrdims + value_types[value_type]
        raw_coo = symbol[:, list(range(nrdims)) + [col_index]]
        # Warning: gams2numpy pkg convert EPS to INF as 5e+300
        if inf_to_zero:
            EPS = raw_coo[:, nrdims] == np.float64("5e+300")
            raw_coo[EPS, nrdims] = 0.0
            if verbose:
                if sum(EPS) > 0:
                    print('GAMS EPS to 0.0 changed')
        value = raw_coo[:, nrdims].astype(float)
        index = {dim: raw_coo[:,idx] for idx, dim in enumerate(metadata['dims'])}
        coords = {dim: metadata['coords'][dim] for dim in metadata['dims']}
        return {'data': (index, value),'coords': coords}

def _symbols_list_from_gdx(filename: str = None, gams_dir: str = None):
    """ It returns a list of symbols' names contained in the GDX file

    Args:
        gams_dir (str, optional): GAMS.exe path, if None the API looks at environment variables. Defaults to None.
        filename (str, optional): GDX filename. Defaults to None.

    Raises:
        Exception: GDX file does not exist or is failed

    Returns:
        list: a list of symbol's names contained in the GDX file
    """
    try:
        from gdxcc import (
            gdxSystemInfo,
            gdxSymbolInfo,
            gdxCreateD,
            gdxOpenRead,
            gdxDataReadDone,
            new_gdxHandle_tp,
            gdxClose,
            gdxFree,
            GMS_SSSIZE,
        )
    except:
        from gams.core.gdx import (
            gdxSystemInfo,
            gdxSymbolInfo,
            gdxCreateD,
            gdxOpenRead,
            gdxDataReadDone,
            new_gdxHandle_tp,
            gdxClose,
            gdxFree,
            GMS_SSSIZE,
        )

    gdxHandle = new_gdxHandle_tp()
    gdxCreateD(gdxHandle, gams_dir, GMS_SSSIZE)
    gdxOpenRead(gdxHandle, filename)
    exists, nSymb, nElem = gdxSystemInfo(gdxHandle)
    symbols = []
    for symNr in range(nSymb):
        ret, name, nrdims, symb_type = gdxSymbolInfo(gdxHandle, symNr)
        symbols.append((name, symb_type, nrdims))
    gdxDataReadDone(gdxHandle)
    gdxClose(gdxHandle)
    gdxFree(gdxHandle)
    return symbols

def _gdx_get_symbol_array_str(symbol_name: str, gdx_file: str,  gams_dir: str=None):
    try:
        from gams2numpy import Gams2Numpy
    except:
        from gams.numpy import Gams2Numpy

    g2np = Gams2Numpy(gams_dir)
    uel_map = g2np.gdxGetUelList(gdx_file)
    arr = g2np.gdxReadSymbolStr(gdx_file, symbol_name,uel_map)
    return arr

def _gdx_get_symbol_data_dict(symbol_name: str, gdx_file: str, gams_dir: str=None):

    try:
        from gdxcc import (
            gdxSymbolInfo,
            gdxFindSymbol,
            gdxSymbolGetDomainX,
            gdxSymbolInfoX,
            gdxCreateD,
            gdxOpenRead,
            new_gdxHandle_tp,
            gdxClose,
            gdxFree,
            GMS_SSSIZE,
        )
    except:
        from gams.core.gdx import (
            gdxSymbolInfo,
            gdxFindSymbol,
            gdxSymbolGetDomainX,
            gdxSymbolInfoX,
            gdxCreateD,
            gdxOpenRead,
            new_gdxHandle_tp,
            gdxClose,
            gdxFree,
            GMS_SSSIZE,
        )

    gdxHandle = new_gdxHandle_tp()
    ret, msg = gdxCreateD(gdxHandle, gams_dir, GMS_SSSIZE)
    ret, msg = gdxOpenRead(gdxHandle, gdx_file)
    assert ret, f"Failed to open '{gdx_file}'"
    ret, symidx = gdxFindSymbol(gdxHandle, symbol_name)
    assert ret, f"Symbol {symbol_name} not found in {gdx_file}"
    if not ret:
        return None
    _, name, NrDims, data_type = gdxSymbolInfo(gdxHandle, symidx)
    _, gdx_domain = gdxSymbolGetDomainX(gdxHandle, symidx)
    _, NrRecs, _, description = gdxSymbolInfoX(gdxHandle, symidx)
    gdxClose(gdxHandle)
    gdxFree(gdxHandle)

    data = {}
    data['symbol'] = symbol_name
    data['dims'] = gdx_domain
    data['coords'] = {dim: sorted(np.sort(_gdx_get_symbol_array_str(symbol_name=dim, gdx_file=gdx_file, gams_dir=gams_dir)[:,0])) for dim in gdx_domain}
    return data



def find_gams():
    if platform.system() == "Linux":
        path = subprocess.check_output(['which', 'gams']).decode().rstrip('\n')
        if path:
            gams_dir_path = os.path.dirname(path)
            exists = os.path.exists(gams_dir_path)
            return exists, gams_dir_path
    elif platform.system() == "Windows":
        path = subprocess.check_output(["where", "gams"]).decode().rstrip('\n')
        if path:
            gams_dir_path = os.path.dirname(path)
            exists = os.path.exists(gams_dir_path)
            return exists, gams_dir_path
    elif platform.system() == "Darwin":
        path = subprocess.check_output(['which', 'gams']).decode().rstrip('\n')
        if path:
            gams_dir_path = os.path.dirname(path)
            exists = os.path.exists(gams_dir_path)
            return exists, gams_dir_path
    return False, None

def exists_gams(gams_dir):
    if gams_dir is None:
        return False, None
    else:
        if os.path.exists(gams_dir):
            path = os.path.join(gams_dir, 'gams')
            try:
                script = subprocess.check_output([path]).decode().rstrip('\n')
                return True, gams_dir
            except FileNotFoundError:
                return False, gams_dir
        else:
            return False, gams_dir
        
def get_gams_dir(gams_dir):
    if gams_dir is None:
        ret, gams_dir = find_gams()
        if ret:
            print(f"Found gams in {gams_dir}")
            return ret, gams_dir
        else:
            print(f'''
                  gams cannot be detected. Make sure you have gams installed. 
                  If gams is already installed, add gams_dir argument with the correct path to gams 
                  e.g. data_collection_instance.adquire(gams_dir="gams/path/here"); or
                  add gams directory to the PATH.
                  ''')
            return ret, gams_dir
    else:
        ret, gams_dir = exists_gams(gams_dir)
        if ret:
            return ret, gams_dir
        else:
            old_gams_dir = gams_dir
            ret, gams_dir = find_gams()
            if ret:
                print(f"Given gams_dir '{old_gams_dir}' was not found. Trying gams in {gams_dir}")
                return ret, gams_dir
            else:
                print(f'''
                    gams was not found in the provided gams_dir {gams_dir}. 
                    Make sure you have gams installed. 
                    If gams is already installed, add gams_dir argument with the correct path to gams 
                    e.g. data_collection_instance.adquire(gams_dir="gams/path/here"); or
                    add gams directory to the PATH.
                    ''')
                return ret, gams_dir




import glob
import json
import os
import shutil
import pathlib
import zipfile
from zipfile import ZIP_DEFLATED, ZIP_LZMA, ZIP_BZIP2

EXT_CONF = {'7z':ZIP_LZMA, 'zip': ZIP_DEFLATED, 'bz2':ZIP_BZIP2}

def _get_files_path_list_from_zip(folder: str, ext: str, zip_ext: str='7z'):
    zip_file_list = glob.glob(os.path.join(folder,f'*.{zip_ext}'))
    files_path = []
    for zip_path in zip_file_list:
        with zipfile.ZipFile(zip_path, mode="r") as zfile:
            for fpath in zfile.namelist():
                complete_path = os.path.join(zip_path,fpath)
                if f".{ext}" in fpath:
                    files_path.append(complete_path)
    return sorted(files_path)

def _get_files_path_list_from_folder(folder: str, ext='*'):
    return glob.glob(os.path.join(folder,f'*/*.{ext}'))

def _get_files_path_list(folder:str, zip_extension:str, file_extension:str):
    if zip_extension is not None:
        file_list = _get_files_path_list_from_zip(folder=folder, ext=file_extension, zip_ext=zip_extension)
        assert len(file_list) > 0, f"No files found in {folder} or in its .{zip_extension} files"
    else:
        file_list = _get_files_path_list_from_folder(folder=folder, ext=file_extension)
        assert len(file_list) > 0, f"No files found in {folder}"
    return sorted(file_list)

def load_scenario_info(path, serializer='json', zip_extension=None):
    '''
    Load scenario info from file.
    '''
    options = ['json','yml']
    assert serializer in options, f"Serializer options are {options}. Provided: '{serializer}'."
    if zip_extension is not None:
        path_parts = path.split(zip_extension+os.sep)
        zip_fpath = path_parts[0] + zip_extension
        target_fpath = path_parts[1]
        with zipfile.ZipFile(zip_fpath, mode="r") as zip_io:
            with zip_io.open(target_fpath) as the_file:
                if serializer == 'json':
                    info = _load_scenario_info_json(the_file)
                elif serializer == 'yml':
                    info = _load_scenario_info_yaml(the_file)
        return info
    else:
        if serializer == 'json':
            return _load_scenario_info_json(path)
        elif serializer == 'yml':
            return _load_scenario_info_yaml(path)

def _load_scenario_info_json(path):
    '''
    Load scenario info from json file.
    '''
    if isinstance(path, str):
        base_name = os.path.basename(path).rstrip('.json')
        with open(path,'r') as stream:
            info = json.load(stream)
            assert all([key in info for key in ['name','metadata']])
            assert info['name'] == base_name, f"Folder name '{base_name}' and 'name':'{info['name']}' do not match."
        return info
    elif isinstance(path, zipfile.ZipExtFile):
        return json.load(path)


def _load_scenario_info_yaml(path):
    '''
    Load scenario info from yaml file compatible with dieterpy <= 1.6.0.
    '''
    import yaml
    if isinstance(path, str):
        base_name = os.path.basename(path).rstrip('_config.yml')
        info = {}
        with open(path,'r') as stream:
            info_ = yaml.load(stream,Loader=yaml.FullLoader)
            info['name'] = info_['id']
            info['metadata'] = info_['config']
            assert info['name'] == base_name, f"Folder name '{base_name}' and 'name':'{info['name']}' do not match."
        return info
    elif isinstance(path, zipfile.ZipExtFile):
        info = {}
        info_ = yaml.load(path,Loader=yaml.FullLoader)
        info['name'] = info_['id']
        info['metadata'] = info_['config']
        # assert info['name'] == base_name, f"Folder name '{base_name}' and 'name':'{info['name']}' do not match."
        return info


def _convert_symbol_name_to_tuple(symbol_name: str):
    '''
    Convert symbol name to tuple.
    '''
    symb_list = symbol_name.split('.')
    if len(symb_list) == 1:
        symb_tp = (symb_list[0],'v')
    elif len(symb_list) == 2:
        symb_tp = (symb_list[0], symb_list[1])
    else:
        raise ValueError(f"Symbol name '{symbol_name}' is not valid")
    return symb_tp


def compress_subdirs(folder: str, zip_extension: str, compresslevel: int=9, allowZip64: bool=True, delete=True):
    child_folders = [path for path in sorted(os.listdir(folder)) if os.path.isdir(os.path.join(folder, path))]
    for childir in child_folders:
        zip_folder = os.path.join(folder, childir)
        compress_dir(zip_folder, zip_extension, compresslevel, allowZip64, delete)
    return None


def compress_dir(folder_path: str, zip_extension: str, compresslevel: int=9, allowZip64: bool=True, delete=True):
    assert zip_extension in EXT_CONF, f"{zip_extension=}. Valid options are: {list(EXT_CONF)}"
    directory = pathlib.Path(folder_path)
    with zipfile.ZipFile(f"{folder_path}.{zip_extension}", "w", EXT_CONF[zip_extension], compresslevel=compresslevel, allowZip64=allowZip64) as archive:
        for file_path in directory.rglob("*"):
            archive.write(file_path, arcname=file_path.relative_to(directory))
    if delete:
        try:
            shutil.rmtree(folder_path)
        except OSError as e:
            print(f"Error folder not deleted {folder_path}")
            #TODO: logger with details about the error (permission denied?)
    return None

def unzip(source_filename, dest_dir, delete=False):
    with zipfile.ZipFile(source_filename) as zf:
        zf.extractall(dest_dir)
    if delete:
        try:
            os.remove(source_filename)
        except OSError as e:
            # If it fails, inform the user.
            print("Error: %s - %s." % (e.filename, e.strerror))

def unzip_all(folder, zip_extension='7z', delete=False):
    options = list(EXT_CONF)
    assert zip_extension in options
    zip_files_found = glob.glob(os.path.join(folder, f'*.{zip_extension}'))
    for file_path in zip_files_found:
        dest_dir = os.path.splitext(file_path)[0]
        unzip(file_path, dest_dir, delete)

def select_scenario_files(scenarios_path, zip_extension=None):
    if zip_extension is None:
        file_list = []
        for file in scenarios_path:
            A = os.path.basename(os.path.dirname(file))
            B = os.path.basename(file).split('.')[0]
            if A == B:
                file_list.append(file)
        return file_list
    else:
        file_list = []
        for file in scenarios_path:
            path_parts = file.split(zip_extension+os.sep)
            zip_fpath = path_parts[0] + zip_extension
            target_fpath = path_parts[1]
            A = os.path.basename(zip_fpath).split('.')[0]
            B = os.path.basename(target_fpath).split('.')[0]
            if A == B:
                file_list.append(file)
        return file_list


from setuptools import setup
import os

packages = []
root_dir = os.path.dirname(__file__)
if root_dir:
    os.chdir(root_dir)

with open(os.path.join(root_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

for dirpath, dirnames, filenames in os.walk("symbolx"):
    # Ignore dirnames that start with '.'
    if "__init__.py" in filenames:
        pkg = dirpath.replace(os.path.sep, ".")
        if os.path.altsep:
            pkg = pkg.replace(os.path.altsep, ".")
        packages.append(pkg)

setup(
    name="symbolx",
    version="0.4.1",
    packages=packages,
    author="Carlos Gaete-Morales",
    author_email="cdgaete@gmail.com",
    install_requires=[
        "numpy",
        "pandas",
        "pyarrow",
        "karray >= 0.2.1",
    ],
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
    description=" Symbolx helps collecting several scenarios data such as multidimentional variables and parameters for reporting and visualization wiht help of karray",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering",
    ],
)


import karray as ka
import json
import os
from ..utils import _get_files_path_list, _convert_symbol_name_to_tuple
import zipfile


def symbol_parser_feather(folder: str, symbol_names: list=[], zip_extension=None, **kwargs):
    '''
    Parse all symbols from a folder and returns a dictionary
    '''

    tool_name = kwargs['tool_name'] if 'tool_name' in kwargs else 'dieter.jl'
    symbol_dict_with_value_type = {}
    for symbs in symbol_names:
        symb_tp = _convert_symbol_name_to_tuple(symbs)
        symbol_dict_with_value_type[symb_tp] = None

    file_list = _get_files_path_list(folder=folder, zip_extension=zip_extension, file_extension='feather')

    symbol_list = []
    for file in file_list:
        if zip_extension is not None:
            path_parts = file.split(zip_extension+os.sep)
            zip_fpath = path_parts[0] + zip_extension
            target_fpath = path_parts[1]
            with zipfile.ZipFile(zip_fpath, mode="r") as zip_io:
                with zip_io.open(target_fpath) as the_file:
                    symbol_info = _info_feather(the_file, tool_name)
        else:
            symbol_info = _info_feather(file, tool_name)
        if (symbol_info['symbol_name'], symbol_info['value_type']) in symbol_dict_with_value_type if len(symbol_dict_with_value_type) != 0 else True:
            symbol_dict = {}
            # This fields are mandatory for a parser
            symbol_dict['symbol_name'] = symbol_info['symbol_name']
            symbol_dict['value_type']  = symbol_info['value_type']
            symbol_dict['path']        = file
            symbol_dict['scenario_name'] = symbol_info['scenario_name']
            # Until here
            # you can add more (custom) attributes. It must be added also see handler.py def add_custom_attr() and be an attribute for loader
            symbol_dict['zip_extension'] = zip_extension
            symbol_list.append(symbol_dict)
    return symbol_list

def load_feather(path:str, zip_extension, **kwargs):
    '''
    Load custom feather file.
    '''
    if "use_threads" in kwargs:
        threads = kwargs["use_threads"]
    else:
        threads = True
    if "with_" in kwargs:
        with_ = kwargs["with_"]
    else:
        with_ = "pandas"
    if zip_extension is not None:
        path_parts = path.split(zip_extension+os.sep)
        zip_fpath = path_parts[0] + zip_extension
        target_fpath = path_parts[1]
        with zipfile.ZipFile(zip_fpath, mode="r") as zip_io:
            with zip_io.open(target_fpath) as the_file:
                arr = ka._from_feather(the_file, use_threads=threads, with_=with_)
        return arr
    else:
        return ka._from_feather(path, use_threads=threads, with_=with_)

def _info_feather(path:str, tool_name:str):
    '''
    Load symbol info from feather file.
    '''
    import pyarrow.feather as ft
    table = ft.read_table(path)
    meta_bstring = table.schema.metadata
    assert tool_name.encode() in meta_bstring, f"{tool_name=} does not exist as key for file metadata {list(meta_bstring)}"
    restored_meta_json = meta_bstring[tool_name.encode()]
    restored_meta = json.loads(restored_meta_json)
    symbol_name = restored_meta['symbol_name']
    value_type = restored_meta['value_type']
    scenario_id = restored_meta['scenario_name']
    return {'symbol_name': symbol_name, 'value_type': value_type, 'scenario_name': scenario_id}
